/* 
 * Tenshi no Hashigo for X
 * Copyright (c) 2001, 2002 TF <tf@denpa.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *      This product includes software developed by TF.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: xgui.c,v 1.8 2002/03/03 15:37:01 tf Exp $
 */
#include <signal.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xlocale.h>
#include <X11/Shell.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeLine.h>
#include <X11/Xaw/SmeBSB.h>

#include "../aladder.h"
#include "xladder.h"

#include "XLCore.h"

static X11Data xdata;

typedef struct {
    Bool screen_mode;
} OptionValue;

static XrmOptionDescRec options[] = {
#ifdef XVIDMODE
    {"-screen", "*screenMode", XrmoptionNoArg, (XtPointer) "True"},
#endif
};

static String fallbacks[] = {
    "*alcore.width: 640",
    "*alcore.height: 480",
    "*alcore.sfont: 12x24",
    "*alcore.mfont: kanji24",
    "*name.international: False",
    NULL,
};

static OptionValue option_value;

static XtResource resources[] = {
#ifdef XVIDMODE
    {"screenMode", "ScreenMode", XtRBool, sizeof(Bool),
     offsetof(OptionValue, screen_mode), XtRImmediate, (XtPointer) False}
    ,
#endif
};

static void MenuPopup(Widget widget, XEvent *e, String *p, Cardinal *n);
static void quit_proc(Widget w, XEvent *e, String *p, Cardinal *n);
static void wm_quit_proc(Widget w, XEvent *e, String *p, Cardinal *n);

static XtActionsRec actions[] = {
    {"wm_quit", (XtActionProc) wm_quit_proc},
    {"quit", (XtActionProc) quit_proc},
    {"menupopup", (XtActionProc) MenuPopup},
};

String translations = "<Message>WM_PROTOCOLS: wm_quit()\n";

String translations_core = "Alt<Key>q: quit()\n" "<Btn2Down>: menupopup()";

Atom wm_protocols[2];

static void
signal_handler(int sig)
{
    XtAppSetExitFlag(xdata.app);
}

static void
quit_proc(Widget w, XEvent *e, String *p, Cardinal *n)
{
    dprintf((stderr, "[λ]\n"));
    XtAppSetExitFlag(XtWidgetToApplicationContext(w));
}

static void
wm_quit_proc(Widget w, XEvent *e, String *p, Cardinal *n)
{
    if (e->xclient.data.l[0] == wm_protocols[0] ||
	e->xclient.data.l[0] == wm_protocols[1]) {
	XtAppSetExitFlag(XtWidgetToApplicationContext(w));
    }
}

static void
RestrictWindowSize(Widget w)
{
    XSizeHints sh;

    sh.flags = PMinSize | PMaxSize;
    sh.min_width = WIDTH;
    sh.min_height = HEIGHT;
    sh.max_width = WIDTH;
    sh.max_height = HEIGHT;
    XSetWMNormalHints(XtDisplay(w), XtWindow(w), &sh);
}

void
InitX11(int argc, char *argv[])
{
    alCore.DependData = (void *) &xdata;

    XtSetLanguageProc(NULL, NULL, NULL);
    xdata.top =
	XtVaOpenApplication(&xdata.app, APPLICATIONCLASS,
			    options, XtNumber(options), &argc, argv,
			    fallbacks, sessionShellWidgetClass, NULL);

    setup_signal(SIGINT, signal_handler);
    setup_signal(SIGTERM, signal_handler);

    XtVaGetApplicationResources(xdata.top,
				(XtPointer) & option_value, resources,
				XtNumber(resources), NULL);
    XtAppAddActions(xdata.app, actions, XtNumber(actions));

    XtOverrideTranslations((Widget) xdata.top,
			   XtParseTranslationTable(translations));

    xdata.core =
	XtVaCreateManagedWidget("alcore", xlCoreWidgetClass,
				xdata.top, NULL);
    XtOverrideTranslations((Widget) xdata.core,
			   XtParseTranslationTable(translations_core));

    XLCoreBindRequestTable();

#ifdef XVIDMODE
    SetupVidMode(xdata.top);
#endif

    XtRealizeWidget(xdata.top);

    wm_protocols[0] =
	XInternAtom(XtDisplay(xdata.top), "WM_DELETE_WINDOW", False);
    wm_protocols[1] =
	XInternAtom(XtDisplay(xdata.top), "WM_SAVEYOURSELF", False);
    XSetWMProtocols(XtDisplay(xdata.top),
		    XtWindow(xdata.top), wm_protocols, 2);

    XtMapWidget(xdata.top);
#if XVIDMODE
    if (xdata.vm.enable_screen_mode && option_value.screen_mode) {
	XGrabPointer(XtDisplay(xdata.top),
		     XtWindow(xdata.top), True, 0, GrabModeAsync,
		     GrabModeAsync, XtWindow(xdata.top), None,
		     CurrentTime);
    }
#endif
    RestrictWindowSize(xdata.top);
}

void
DeleteX11()
{
    WindowMode();
    XtUnrealizeWidget(xdata.top);
    XtDestroyWidget(xdata.top);
}

typedef struct {
    char *name;
    int enable;
} MenuItem;

MenuItem menu_items[] = {
    {"save", False},
    {"load", False},
    {"quit", False},
    {NULL, False},
};

/* 
 * ܥ˥塼˴롥
 */
static void
MenuDestroy(Widget w, XtPointer dat, XtPointer call)
{
    XtDestroyWidget((Widget) dat);
}

static void
MenuExecCommand(const char *command)
{
    if (!strcmp(command, "save")) {
	SaveGame("quick");
    } else if (!strcmp(command, "load")) {
	LoadGame("quick");
    } else if (!strcmp(command, "quit")) {
	LongJump(JUMP_END);
    } else {
	fprintf(stderr, "̤ΤΥޥɤǤ");
    }
}

/* 
 * ˥塼¹
 */
static void
MenuExec(Widget w, XtPointer dat, XtPointer call)
{
    MenuExecCommand(XtName(w));
}

/* 
 * ֲ
 */
static void
MenuUngrab(Widget w, XtPointer dat, XtPointer call)
{
    XtUngrabPointer((Widget) dat, CurrentTime);
}

/* 
 * ܥ˥塼롥
 */
static void
MenuPopup(Widget widget, XEvent *e, String *p, Cardinal *n)
{
    Widget menu;

    menu =
	XtVaCreatePopupShell("menu", simpleMenuWidgetClass, xdata.top,
			     NULL);

    {
	MenuItem *p = menu_items;
	XLCoreWidget cw = (XLCoreWidget) widget;

	while (p->name != NULL) {
	    Widget w;
	    w = XtVaCreateManagedWidget(p->name, smeBSBObjectClass, menu,
					NULL);
	    XtAddCallback(w, XtNcallback, MenuExec, (XtPointer) cw);
	    p++;
	}
    }

    XtAddCallback(menu, XtNpopdownCallback, MenuUngrab, xdata.top);
    XtAddCallback(menu, XtNpopdownCallback, MenuDestroy, (XtPointer) menu);

    /* ˥塼ư */
    {
	String param = "menu";
	XtCallActionProc(menu, "XawPositionSimpleMenu", e, &param, 1);
    }

    XtPopupSpringLoaded(menu);

    XtGrabPointer(xdata.top,
		  True, 0,
		  GrabModeAsync, GrabModeAsync,
		  XtWindow(xdata.top), None, CurrentTime);
}
