/* 
 * Tenshi no Hashigo for X
 * Copyright (c) 2001, 2002 TF <tf@denpa.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *      This product includes software developed by TF.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: ximage.c,v 1.9 2002/03/04 18:11:06 tf Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "ximage.h"
#include "../aladder.h"

static struct {
    int id;
    TrueColorMap *tcmap;
} tcmap[10];

static int tcmap_idx = 0;

static int a_map[17][512];	/* alpha factor */
static int i_map[33][256];	/* intensity factor */

static void get_visual_info(Display *, Window, XLImage *,
			    XWindowAttributes);

XLImage *
XLImageCreate(Display *d, Window w, int width, int height)
{
    int major, minor;
    XLImage *image;
    int depth;
    Visual *visual;
    XWindowAttributes att;

    if ((image = (XLImage *) malloc(sizeof(XLImage))) == NULL) {
	perror("XLImageCreate:malloc");
	exit(1);
    }

    image->display = d;
    image->window = w;
    image->xoffset = 0;
    image->yoffset = 0;
    image->rgbmap = NULL;
    image->pixel_num = 0;

    if ((image->support_shmem = XShmQueryExtension(d))) {
	XShmQueryVersion(d, &major, &minor, &image->support_shpix);
	dprintf((stderr, "Support MIT-Shmem Extention %d.%d\n", major,
		minor));
    }
    /* image->support_shmem = False; */

    XGetWindowAttributes(d, w, &att);
    depth = att.depth;
    visual = att.visual;

    image->colormap = att.colormap;
    image->gc = DefaultGC(d, XScreenNumberOfScreen(att.screen));

    /* get visual information */
    get_visual_info(d, w, image, att);

    if (image->support_shmem) {
	dprintf((stderr, "Using shared images.\n"));

	image->ximage = XShmCreateImage(d, visual, depth, ZPixmap, NULL,
					&image->shminfo, width, height);

	if ((image->shminfo.shmid =
	     shmget(IPC_PRIVATE, image->ximage->bytes_per_line * height,
		    IPC_CREAT | 0777)) < 0) {
	    fprintf(stderr, "shmget: size:%x %s\n",
		    image->ximage->bytes_per_line * height,
		    strerror(errno));
	    exit(1);
	}

	if ((image->shminfo.shmaddr = image->ximage->data =
	     shmat(image->shminfo.shmid, 0, 0)) == NULL) {
	    perror("shmat");
	    exit(1);
	}

	image->shminfo.readOnly = False;
	XShmAttach(d, &image->shminfo);

	/* shared pixmap */
	if (image->support_shpix) {
	    dprintf((stderr, "Get shared pixmap\n"));

	    image->pixmap = XShmCreatePixmap(d, w,
					     image->ximage->data,
					     &image->shminfo,
					     image->ximage->width,
					     image->ximage->height,
					     image->ximage->depth);
	} else {
	    dprintf((stderr, "Get normal pixmap\n"));
	    image->pixmap = XCreatePixmap(d, w,
					  image->ximage->width,
					  image->ximage->height,
					  image->ximage->depth);
	}
    } else {
	dprintf((stderr, "Get normal image\n"));
	image->ximage = XCreateImage(d, visual, depth, ZPixmap, 0, NULL,
				     width, height, 8, 0);

	if ((image->ximage->data = malloc(image->ximage->bytes_per_line *
					  image->ximage->height)) ==
	    NULL) {
	    perror("XLImageCreate");
	    exit(1);
	}
	image->pixmap = XCreatePixmap(d, w,
				      image->ximage->width,
				      image->ximage->height,
				      image->ximage->depth);
    }

    return image;
}

static void
get_visual_info(Display *d, Window w, XLImage *image,
		XWindowAttributes att)
{
    int matched;
    XVisualInfo *vinfolist;
    XVisualInfo vinfo;

    /* get visual */
    vinfo.visualid = XVisualIDFromVisual(att.visual);
    vinfolist = XGetVisualInfo(d, VisualIDMask, &vinfo, &matched);
    if (!matched) {
	fprintf(stderr, "Can't get visual info.\n");
	exit(1);
    }
    image->visual_class = vinfolist[0].class;
    XFree(vinfolist);

    if (image->visual_class == TrueColor
	|| image->visual_class == DirectColor) {
	int i;
	TrueColorMap *map;

	dprintf((stderr, "TrueColor mode.\n"));

	for (i = 0; i < tcmap_idx; i++) {
	    if (vinfo.visualid == tcmap[i].id) {
		image->tcmap = tcmap[i].tcmap;
		goto end;
	    }
	}
	if ((map = (TrueColorMap *) malloc(sizeof(TrueColorMap))) == NULL) {
	    perror("get_visual_info: malloc");
	    exit(1);
	}
	/* Get colormap for TrueColor */
	for (i = 0; i < 256; i++) {
	    int col;
	    XColor defs;
	    defs.flags = DoRed | DoGreen | DoBlue;

	    defs.red = col = (i << 8) | i;
	    defs.green = 0;
	    defs.blue = 0;
	    XAllocColor(d, image->colormap, &defs);
	    map->red[i] = defs.pixel;

	    defs.red = 0;
	    defs.green = col;
	    defs.blue = 0;
	    XAllocColor(d, image->colormap, &defs);
	    map->green[i] = defs.pixel;

	    defs.red = 0;
	    defs.green = 0;
	    defs.blue = col;
	    XAllocColor(d, image->colormap, &defs);
	    map->blue[i] = defs.pixel;
	}
	image->tcmap = tcmap[tcmap_idx++].tcmap = map;
    }
  end:
}

void
XLImageDestroy(XLImage *image)
{
    if (image == NULL)
	return;

    if (image->pixel_num > 0) {
	XFreeColors(image->display, image->colormap,
		    image->allocated_pixel, image->pixel_num, 0);
	image->pixel_num = 0;
    }

    XFreePixmap(image->display, image->pixmap);
    if (image->support_shmem) {
	/* detach shared memory */
	XShmDetach(image->display, &image->shminfo);
	image->ximage->data = NULL;
	/* delete shared memory */
	shmdt(image->shminfo.shmaddr);
	shmctl(image->shminfo.shmid, IPC_RMID, NULL);
    }

    /* destroy image */
    XDestroyImage(image->ximage);

    if (image->rgbmap != NULL)
	free(image->rgbmap);

    free(image);
}

void
XLImagePut(XLImage *image)
{
    if (image->support_shmem) {
	if (!image->support_shpix) {
	    /* no shared pixmap */
	    XShmPutImage(image->display, image->pixmap, image->gc,
			 image->ximage, 0, 0, 0, 0,
			 image->ximage->width,
			 image->ximage->height, False);
	} else {
	    /* do nothing, because shared memory works! */
	}
    } else {
	XPutImage(image->display, image->pixmap, image->gc,
		  image->ximage, 0, 0, 0, 0,
		  image->ximage->width, image->ximage->height);
    }
    XCopyArea(image->display, image->pixmap, image->window, image->gc,
	      0, 0, image->ximage->width, image->ximage->height, 0, 0);
}

void
XLImagePutArea(XLImage *image, int x, int y, int width, int height)
{
    if (image->support_shmem) {
	if (!image->support_shpix) {
	    /* no shared pixmap */
	    XShmPutImage(image->display, image->pixmap, image->gc,
			 image->ximage, x, y, x, y, width, height, False);
	} else {
	    /* do nothing, because shared memory works! */
	}
    } else {
	XPutImage(image->display, image->pixmap, image->gc,
		  image->ximage, x, y, x, y, width, height);
    }
    XCopyArea(image->display, image->pixmap, image->window, image->gc,
	      x, y, width, height, x, y);
}

/* 
 * interface functions
 */
#define SXIMAGE_TPIXEL_SCREEN(s,r,g,b)\
(s->tcmap->red[r]|s->tcmap->green[g]|s->tcmap->blue[b])

void
set_ximage(Image *image, XLImage *xlimage, int width, int height)
{
    int i, j, idx;
    long pixel;
    int r, g, b;
    u_char *p = (u_char *)image->bitmap;

    switch (xlimage->visual_class) {
    case TrueColor:
    case DirectColor:
	for (i = 0; i < image->height; i++) {
	    for (j = 0; j < image->width; j++) {
		idx = (j + image->width * i) * 3;
		b = p[idx];
		g = p[idx + 1];
		r = p[idx + 2];
		pixel = SXIMAGE_TPIXEL_SCREEN(xlimage, r, g, b);
		XPutPixel(xlimage->ximage, j, i, pixel);
	    }
	}
	break;
    default:
	dprintf((stderr, "set_ximage: Visual not supported.\n"));
	exit(1);
	break;
    }
}

void
set_ximage_area(Image *image, XLImage *xlimage, int x, int y, int width,
		int height)
{
    int i, j, idx;
    long pixel;
    int r, g, b;
    u_char *p = (u_char *)image->bitmap;

    switch (xlimage->visual_class) {
    case TrueColor:
    case DirectColor:
	for (i = 0; i < height; i++) {
	    for (j = 0; j < width; j++) {
		idx = (x + j + image->width * (y + i)) * 3;
		b = p[idx];
		g = p[idx + 1];
		r = p[idx + 2];
		pixel = SXIMAGE_TPIXEL_SCREEN(xlimage, r, g, b);
		XPutPixel(xlimage->ximage, x + j, y + i, pixel);
	    }
	}
	break;
    default:
	fprintf(stderr, "set_ximage: Visual not supported.\n");
	exit(1);
	break;
    }
}

void
InitXLImageProcTable()
{
    int i, j;

    /* alpha map */
    for (i = 0; i < 17; i++) {
	for (j = -255; j < 256; j++) {
	    a_map[i][j + 255] = i * j / 16;
	}
    }
    /* intensity map */
    for (i = 0; i < 33; i++) {
	for (j = 0; j < 256; j++) {
	    i_map[i][j] = i * j / 32;
	}
    }
}
