/* 
 * Tenshi no Hashigo for X
 * Copyright (c) 2001, 2002 TF <tf@denpa.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgment:
 *      This product includes software developed by TF.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: image.c,v 1.34 2002/03/05 17:17:14 tf Exp $
 */

#include <stdlib.h>

#include "aladder.h"

void text_coordinate(int x, int y, int *tx, int *ty);
void graphics_coordinate(int x, int y, int *tx, int *ty);

static void ImageDestroy(Image *image);
static void ImageAdd(Image *dst, Image *src, int xoff, int yoff);

/* 
 * åϢ(LRU)
 */
#define IMAGE_CACHE_SIZE 8

typedef struct cache {
    char name[ARG_LEN];
    Image *image;

    struct cache *prev;
    struct cache *next;
} image_cache_t;

static image_cache_t cache[IMAGE_CACHE_SIZE];
static image_cache_t cache_list;

void font_cache_new();

static void
image_cache_insert(image_cache_t *pos, image_cache_t *x)
{
    pos->prev->next = x;
    x->prev = pos->prev;
    x->next = pos;
    pos->prev = x;
}

static image_cache_t *
image_cache_remove(image_cache_t *x)
{
    x->prev->next = x->next;
    x->next->prev = x->prev;

    return x;
}

void
image_cache_new()
{
    int i;

    cache_list.next = cache_list.prev = &cache_list;
    for (i = 0; i < IMAGE_CACHE_SIZE; i++) {
	cache[i].image = NULL;
	image_cache_insert(&cache_list, &cache[i]);
    }
}

#if 0
static void
image_cache_delete()
{
    int i;

    for (i = 0; i < IMAGE_CACHE_SIZE; i++)
	if (cache[i].image != NULL)
	    ImageDestroy(cache[i].image);
}
#endif

static void
dup_image(Image *dst, Image *src)
{
    if (dst->width != src->width || dst->height != src->height) {
	if (dst->bitmap)
	    free(dst->bitmap);
	{
	    int w = src->width * 3 + src->width % 4;
	    dst->bitmap =
		(u_char *) calloc(w * src->height, sizeof(u_char));
	}
    }
    dst->width = src->width;
    dst->height = src->height;
    dst->name = src->name;
    ImageAdd(dst, src, 0, 0);
}

/* 
 * 
 */
Color
get_color(u_char r, u_char g, u_char b)
{
    Color ret;
    ret.r = r;
    ret.g = g;
    ret.b = b;

    return ret;
}

/* 
 * Image ؿ
 */
static Image *
ImageCreate(int width, int height)
{
    Image *img;

    img = (Image *) malloc(sizeof(Image));
    img->width = width;
    img->height = height;
    if (width * height) {
	int w = width * 3 + width % 4;
	img->bitmap = (u_char *) calloc(w * height, sizeof(u_char));
    } else {
	img->bitmap = NULL;
    }

    return img;
}

static void
ImageDestroy(Image *image)
{
    if (image) {
	if (image->bitmap)
	    free(image->bitmap);
	free(image);
    }
}

static void
ImageLoad(Image *image, const char *name)
{
    image_cache_t *p;

    for (p = cache_list.next; p != &cache_list; p = p->next) {
	if (!strcmp(name, p->name)) {
	    image_cache_remove(p);
	    image_cache_insert(&cache_list, p);
	    dup_image(image, p->image);
	    return;
	}
    }

    p = image_cache_remove(cache_list.next);
    image_cache_insert(&cache_list, p);

    if (p->image != NULL)
	ImageDestroy(p->image);
    p->image = ImageCreate(image->width, image->height);
    strcpy(p->name, name);
    get_image(alCore.cg_arc, name, p->image);
    dup_image(image, p->image);
}

static void
ImageFill(Image *image, int x, int y, int w, int h, Color color)
{
    int i, j, idx;
    u_char *p = (u_char *) image->bitmap;

    for (i = 0; i < h; i++) {
	idx = (image->width * (y + i) + x) * 3;
	for (j = 0; j < w; j++) {
	    p[idx++] = color.r;
	    p[idx++] = color.g;
	    p[idx++] = color.b;
	}
    }
}

static void
ImageAdd(Image *dst, Image *src, int xoff, int yoff)
{
    u_char *si = (u_char *) src->bitmap;
    u_char *di = (u_char *) dst->bitmap;

    if (src->width == dst->width && src->height == dst->height &&
	xoff == 0 && yoff == 0) {
	memcpy(dst->bitmap, src->bitmap, src->width * src->height * 3);
    } else {
	int i, j, hs, vs, hd, vd, w, h;

	hs = vs = hd = vd = 0;
	w = (src->width < dst->width) ? src->width : dst->width;
	w = (xoff > 0) ? w - xoff : w + xoff;
	h = (src->height < dst->height) ? src->height : dst->height;
	h = (yoff > 0) ? h - yoff : h + yoff;

	if (xoff > 0) {
	    hs = 0;
	    hd = xoff;
	}
	if (xoff < 0) {
	    hs = -xoff;
	    hd = 0;
	}
	if (yoff > 0) {
	    vs = 0;
	    vd = yoff;
	}
	if (yoff < 0) {
	    vs = -yoff;
	    vd = 0;
	}
	for (i = 0; i < h; i++) {
	    u_char *p =
		di + (dst->width * 3 + dst->width % 4) * (i + vd) + hd * 3;
	    u_char *q =
		si + (src->width * 3 + src->width % 4) * (i + vs) + hs * 3;
	    for (j = 0; j < w * 3; j++) {
		*p++ = *q++;
	    }
	}
    }
}

static Image *
ImageCreateMask(Image *src)
{
    int i, j;
    u_char r, g, b;
    u_char *si = (u_char *) src->bitmap;
    Image *dst = ImageCreate(src->width, src->height);
    u_char *di = (u_char *) dst->bitmap;

    r = si[0];
    g = si[1];
    b = si[2];

    for (i = 0; i < src->height; i++) {
	u_char *p = di + (dst->width * 3 + dst->width % 4) * i;
	u_char *q = si + (src->width * 3 + src->width % 4) * i;
	for (j = 0; j < src->width; j++, p += 3, q += 3) {
	    if (*q == r && *(q + 1) == g && *(q + 2) == b) {
		*p = 0xff;
		*(p + 1) = 0xff;
		*(p + 2) = 0xff;
	    } else {
		*p = 0x0;
		*(p + 1) = 0x0;
		*(p + 2) = 0x0;
	    }
	}
    }
    return dst;
}

static void
ImageCopy(Image *dst, Image *src, int xd, int yd, int xs, int ys, int w,
	  int h)
{
    int i, j, dwidth, swidth;
    u_char *si = (u_char *) src->bitmap;
    u_char *di = (u_char *) dst->bitmap;

    dwidth = dst->width * 3 + dst->width % 4;
    swidth = src->width * 3 + src->width % 4;

    xd *= 3;
    xs *= 3;
    w *= 3;

    for (i = 0; i < h; i++) {
	u_char *p = di + dwidth * (i + yd) + xd;
	u_char *q = si + swidth * (i + ys) + xs;
	for (j = 0; j < w; j++) {
	    *p++ = *q++;
	}
    }
}

static void
ImageMerge(Image *dst, Image *src1, Image *src2, int alpha)
{
    int i, j;
    int dwidth, s1width, s2width;
    u_char *s1i = (u_char *) src1->bitmap;
    u_char *s2i = (u_char *) src2->bitmap;
    u_char *di = (u_char *) dst->bitmap;

    alpha = (alpha > 0xff) ? 0xff : alpha;

    if (dst->width != src1->width || dst->width != src2->width ||
	dst->height != src1->height || dst->height != src2->height) {
	fprintf(stderr, "ƱβޡǤޤ\n");
	return;
    }

    dwidth = dst->width * 3 + dst->width % 4;
    s1width = src1->width * 3 + src1->width % 4;
    s2width = src2->width * 3 + src2->width % 4;

    for (i = 0; i < src1->height; i++) {
	u_char *p = di + dwidth * i;
	u_char *q = s1i + s1width * i;
	u_char *r = s2i + s2width * i;
	for (j = 0; j < src1->width * 3; j++) {
	    *p++ = alpha_map[*q++][*r++][alpha];
	}
    }
}

static void
ImageAddWithMask(Image *dst, Image *src, Image *mask, int xoff, int yoff)
{
    int i, j;
    int sxoff = 0, syoff = 0;
    int swidth, dwidth, mwidth;
    int width = src->width;
    int height = src->height;
    u_char *si = (u_char *) src->bitmap;
    u_char *di = (u_char *) dst->bitmap;
    u_char *mi = (u_char *) mask->bitmap;

    if (xoff < 0) {
	sxoff = -xoff;
	width += xoff;
	xoff = 0;
    }

    if (yoff < 0) {
	syoff = -yoff;
    }

    if (xoff + width > dst->width)
	width -= xoff + width - dst->width;
    if (yoff + height > dst->height)
	height -= yoff + height - dst->height;

    swidth = src->width * 3 + src->width % 4;
    dwidth = dst->width * 3 + dst->width % 4;
    mwidth = mask->width * 3 + mask->width % 4;

    width *= 3;
    xoff *= 3;
    sxoff *= 3;

    for (i = syoff; i < height; i++) {
	u_char *p = di + dwidth * (i + yoff) + xoff;
	u_char *q = si + swidth * i + sxoff;
	u_char *r = mi + mwidth * i + sxoff;
	for (j = 0; j < width; j++, p++, q++, r++) {
	    *p = alpha_map[*p][*q][*r];
	}
    }
}

static void
ImageDisplay(int brightness)
{
    u_char *vi = (u_char *) alCore.vram->bitmap;
    u_char *wi = (u_char *) alCore.work->bitmap;

    if (TEXT.fullscreen || CURRENT_STATE == STAT_BACKLOG) {
	if (brightness == 20) {
	    memcpy(vi, wi, WIDTH * HEIGHT * 3);
	} else {
	    int i;
	    for (i = 0; i < WIDTH * HEIGHT * 3; i++) {
		vi[i] = bri_map[brightness][wi[i]];
	    }
	}
    } else {
	int i, j, idx;
	memcpy(vi, wi, WIDTH * HEIGHT * 3);

	idx = TEXT_WINDOW_TOP * WIDTH * 3;
	for (i = TEXT_WINDOW_TOP; i < HEIGHT; i++) {
	    for (j = 0; j < WIDTH * 3; j++, idx++)
		vi[idx] = bri_map[brightness][wi[idx]];
	}

	for (i = TEXT_WINDOW_BORDER_TOP; i < TEXT_WINDOW_TOP; i++) {
	    int bri =
		brightness +
		((i - TEXT_WINDOW_TOP) * (20 -
					  brightness)) /
		(TEXT_WINDOW_BORDER_TOP - TEXT_WINDOW_TOP);
	    idx = i * WIDTH * 3;
	    for (j = 0; j < WIDTH * 3; j++) {
		vi[idx + j] = bri_map[bri][wi[idx + j]];
	    }
	}
    }
}

static void
CharacterLoad(Image *dst, char arg[][ARG_LEN], int num)
{
    Image *img, *mask;
    char mask_name[ARG_LEN];
    int xoff, cmd;

    img = ImageCreate(0, 0);
    mask = ImageCreate(0, 0);

    sprintf(mask_name, "m%s", arg[1]);

    ImageLoad(img, arg[1]);
    ImageLoad(mask, mask_name);

    cmd = strtol(arg[3], NULL, 0);

    if (!strcmp(arg[2], "L")) {
	xoff = CHARA_LEFT_MARGIN;
	strcpy(SCENE.left, arg[1]);
    } else if (!strcmp(arg[2], "M")) {
	xoff = WIDTH / 2 - img->width / 2;
	strcpy(SCENE.middle, arg[1]);
    } else if (!strcmp(arg[2], "R")) {
	xoff = WIDTH - img->width - CHARA_RIGHT_MARGIN;
	strcpy(SCENE.right, arg[1]);
    } else {
	dprintf((stderr, "Ωɽֻ꤬Ǥ[%s]\n", arg[2]));
	/* 
	 * "LMR" ǤʤդĤ(ץȤΥХ?)
	 */
	xoff = WIDTH / 2 - img->width / 2;
	strcpy(SCENE.middle, arg[1]);
    }
    if (cmd >= 0)
	ImageAddWithMask(dst, img, mask, xoff, HEIGHT - img->height);

    ImageDestroy(img);
    ImageDestroy(mask);
}

void
ImageBox(Image *image, int x, int y, int w, int h, Color color)
{
    int i, j, idx;
    u_char *p = (u_char *) image->bitmap;

    idx = (image->width * y + x) * 3;
    for (j = 0; j < w; j++) {
	p[idx++] = color.b;
	p[idx++] = color.g;
	p[idx++] = color.r;
    }

    for (i = 0; i < h; i++) {
	idx = (image->width * (y + i) + x) * 3;
	p[idx++] = color.b;
	p[idx++] = color.g;
	p[idx++] = color.r;

	idx = (image->width * (y + i) + x + w) * 3;
	p[idx++] = color.b;
	p[idx++] = color.g;
	p[idx++] = color.r;
    }

    idx = (image->width * (y + h) + x) * 3;
    for (j = 0; j < w; j++) {
	p[idx++] = color.b;
	p[idx++] = color.g;
	p[idx++] = color.r;
    }
}

#ifdef FAST_EFFECT
static int fade_mask[8][4][4] = {
    {{0, 0, 1, 0},
     {1, 0, 0, 0},
     {0, 0, 0, 0},
     {0, 0, 0, 0}},
    {{1, 0, 1, 0},
     {0, 0, 0, 0},
     {1, 0, 1, 0},
     {0, 0, 0, 0}},
    {{1, 0, 1, 0},
     {0, 0, 0, 1},
     {1, 0, 1, 0},
     {0, 1, 0, 0}},
    {{1, 0, 1, 0},
     {0, 1, 0, 1},
     {1, 0, 1, 0},
     {0, 1, 0, 1}},
    {{1, 1, 1, 0},
     {0, 1, 0, 1},
     {1, 0, 1, 1},
     {0, 1, 0, 1}},
    {{1, 1, 1, 1},
     {0, 1, 0, 1},
     {1, 1, 1, 1},
     {0, 1, 0, 1}},
    {{1, 1, 1, 1},
     {1, 1, 0, 1},
     {1, 1, 1, 1},
     {0, 1, 1, 1}},
    {{1, 1, 1, 1},
     {1, 1, 1, 1},
     {1, 1, 1, 1},
     {1, 1, 1, 1}},
};

void
ImageBlendPattern8(Image *dst, Image *src, int x, int y, int step)
{
    int i, j, w;
    u_char *di = (u_char *) dst->bitmap;
    u_char *si = (u_char *) src->bitmap;

    /* src  dst βƱȤ */
    w = dst->width * 3 + dst->width % 4;

    for (i = 0; i < 16; i++) {
	u_char *p = di + w * (y + i) + x * 3;
	u_char *q = si + w * (y + i) + x * 3;

	for (j = 0; j < 16; j++, p += 3, q += 3) {
	    if (fade_mask[step][i % 4][j % 4]) {
		*p = *q;
		*(p + 1) = *(q + 1);
		*(p + 2) = *(q + 2);
	    }
	}
    }
}
#endif

void
ImageBlendPattern(Image *dst, Image *src, int x, int y, int step)
{
    static int xoff[] = { 3, 1, 1, 3, 2, 0, 0, 2, 3, 1, 1, 3, 2, 0, 0, 2 };
    static int yoff[] = { 0, 2, 0, 2, 1, 3, 1, 3, 1, 3, 1, 3, 0, 2, 0, 2 };
    int w, i, j;
    u_char *p, *q;
    u_char *di = (u_char *) dst->bitmap;
    u_char *si = (u_char *) src->bitmap;

    /* src  dst βƱ 4 ܿǤ뤳Ȥ */
    w = dst->width * 3 + dst->width % 4;

    for (i = 0; i < 4; i++) {
	for (j = 0; j < 4; j++) {
	    p = di + w * (y + yoff[step] + 4 * i) + (x + xoff[step] +
						     4 * j) * 3;
	    q = si + w * (y + yoff[step] + 4 * i) + (x + xoff[step] +
						     4 * j) * 3;
	    *p = *q;
	    *(p + 1) = *(q + 1);
	    *(p + 2) = *(q + 2);
	}
    }
}

void
ImageDarken(Image *image, int brightness)
{
    int i;
    u_char *p = (u_char *) image->bitmap;
    int size = (image->width * 3 + image->width % 4) * image->height;

    for (i = 0; i < size; i++)
	p[i] = bri_map[brightness][p[i]];
}

void
ImageHalfSize(Image *dst, Image *src)
{
    int i, j, dwidth, swidth;
    u_char *di = (u_char *) dst->bitmap;
    u_char *si = (u_char *) src->bitmap;

    dwidth = dst->width * 3 + dst->width % 4;
    swidth = src->width * 3 + src->width % 4;

    for (i = 0; i < dst->height; i++) {
	u_char *p = di + dwidth * i;
	u_char *q = si + swidth * i * 2;
	for (j = 0; j < dst->width; j++, p += 3, q += 6) {
	    *p = *q;
	    *(p + 1) = *(q + 1);
	    *(p + 2) = *(q + 2);
	}
    }
}

void
ImageLibInit()
{
    int i, j, k;

    for (i = 0; i <= 20; i++) {
	for (j = 0; j < 256; j++) {
	    int r = j * i / 20;
	    bri_map[i][j] = (u_char) r;
	}
    }

    for (i = 0; i < 256; i++) {
	for (j = 0; j < 256; j++) {
	    for (k = 0; k < 256; k++) {
		int r = (i * k + j * (0xff - k)) / 0xff;
		alpha_map[i][j][k] = (u_char) r;
	    }
	}
    }

    BLACK = get_color(0x0, 0x0, 0x0);
    WHITE = get_color(0xff, 0xff, 0xff);
    GRAY = get_color(0xbe, 0xbe, 0xbe);
    YELLOW = get_color(0xff, 0xff, 0x0);
    PINK = get_color(0xff, 0xc0, 0xc0);

    /* API ΥХ */
    ALIB.imageNew = ImageCreate;
    ALIB.imageDelete = ImageDestroy;
    ALIB.imageLoad = ImageLoad;
    ALIB.imageFill = ImageFill;
    ALIB.imageAdd = ImageAdd;
    ALIB.imageCreateMask = ImageCreateMask;
    ALIB.imageCopy = ImageCopy;
    ALIB.imageAddWithMask = ImageAddWithMask;
    ALIB.imageMerge = ImageMerge;
    ALIB.imageDisplay = ImageDisplay;
    ALIB.imageCharacterLoad = CharacterLoad;
    ALIB.imageBox = ImageBox;
    ALIB.imageHalfSize = ImageHalfSize;
    ALIB.imageBlendPattern = ImageBlendPattern;
    ALIB.imageDarken = ImageDarken;

    ALIB.putChar = PutChar;
    ALIB.drawImageString = DrawImageString;
    ALIB.imageStringWidth = ImageStringWidth;
    ALIB.coordTextToGraphics = graphics_coordinate;
    ALIB.coordGraphicsToText = text_coordinate;

    image_cache_new();
    font_cache_new();		/* եȥåκ */
}
